/*
 * DisjointUnion.java
 *
 * Created on February 18, 2002, 12:03 PM
 */

package org.semanticweb.dl;

import java.util.*;
import java.io.*;


/**
 * DisjointUnion handles daml:disjointUnionOf. 
 * @author  Michael Sintek
 * @version 
 */
public class DisjointUnion extends Expression {

  ArrayList _expressions; // of Expression
  
  /** Creates new DisjointUnion */
  public DisjointUnion (ArrayList expressions) {
    _expressions = expressions;
  }
  
  public void print (Syntax syntax, PrintWriter out) {
    if (syntax.XML ()) {
      out.print ("<DISJOINT_UNION>");
      printSequence (_expressions, syntax, out);
      out.print ("</DISJOINT_UNION>");
    } else {
      out.print ("(disjoint-union ");
      printSequence (_expressions, syntax, out);
      out.print (")");
    }
  }
  
  public Expression toSHIQ(KB kb, int mode) { // we need access to the kb !!
    // (disjoint-union c1 c2 ...) -> 
    // (disjoint c1 c2 ...) & return (or c1 c2 ...)
    // disjoint -> add pairwise (implies ci (not cj)) to kb
    // common part: list of SHIQ expressions
    ArrayList list = new ArrayList();
    for (Iterator iter = _expressions.iterator(); iter.hasNext();)
      list.add(((Expression)iter.next()).toSHIQ(kb, CONCEPT));
    // disjoint part
    int l = list.size();
    if (l < 2)
      System.err.println("warning: disjoint union has less than 2 classes: " 
        + this.toString());
    else { // pairwise disjointness
      for (int i = 0; i < l-1; i++) {
        Expression c1 = (Expression)list.get(i);
        for (int j = i+1; j < l; j++) {
          Expression c2 = (Expression)list.get(j);
          // (implies c1 (not c2))
          kb.add(new Implies(c1, new Not(c2)));
        }
      }
    }
    // union part
    return Or.getOr(list);
  } 

}
