/*
 * KB.java
 *
 * Created on February 18, 2002, 9:26 AM
 */

package org.semanticweb.dl;

import java.util.*;
import java.io.*;
import com.hp.hpl.mesa.rdf.jena.model.*;


/**
 * A DL knowledge base.
 * @author  Michael Sintek
 * @version 
 */
public class KB {

  HashMap _symbols; // resource -> Symbol
  ArrayList _assertions; // of Assertion
  HashSet _concepts; // of Symbol
  HashSet _roles; // of Symbol
  HashMap _abbrevs = null; // of namespace -> abbrev (Strings)
  int _gensymCounter = 0;
  
  /** Creates new KB */
  public KB() {
    _symbols = new HashMap();
    _assertions = new ArrayList();
    _concepts = new HashSet();
    _roles = new HashSet();
  }
  
  public void addConcept(Symbol concept) {
    if (!_concepts.contains(concept))
      _concepts.add(concept);
  }
  
  public void addRole(Symbol role) {
    if (!_roles.contains(role))
      _roles.add(role);
  }
  
  public void add(Assertion assertion) {
    _assertions.add(assertion);
    if (assertion instanceof Defconcept)
      addConcept(((Defconcept)assertion).getSymbol());
    else if (assertion instanceof Defconcept)
      addRole(((Defrole)assertion).getSymbol());
  }
  
  public Symbol getSymbol(Resource node) {
    Symbol symbol = (Symbol)_symbols.get(node);
    if (symbol == null) {
      // check for TOP and BOTTOM!!! ... ???
      symbol = new Symbol(node);
      _symbols.put(node, symbol);
    }
    return symbol;
  }

  public boolean isConcept(Symbol symbol) {
    return _concepts.contains(symbol);
  }

  
  // transformation to SHIQ
  
  public KB toSHIQ() {
    // transform this KB to a SHIQ compatible one
    KB newKB = new KB();
    for (Iterator iter = _assertions.iterator(); iter.hasNext();) {
      ((Assertion)iter.next()).toSHIQ(newKB);
    }
    // find missing concepts and roles and add assertions for them
    HashSet concepts = new HashSet();
    HashSet roles = new HashSet();
    for (Iterator iter = newKB._assertions.iterator (); iter.hasNext ();) {
      Assertion assertion = ((Assertion)iter.next ());
      if (assertion instanceof Defconcept)
        concepts.add(((Defconcept)assertion).getSymbol());
      else if (assertion instanceof Defrole)
        roles.add(((Defrole)assertion).getSymbol());
    }
    // roles = new HashSet();
    for (Iterator citer = newKB._concepts.iterator(); citer.hasNext();) {
      Symbol concept = (Symbol)citer.next();
      if (   !concept.isTop() && !concept.isBottom() 
          && !concepts.contains(concept)) { // not yet defined
        System.err.println("Warning: concept not defined: " + concept);
        newKB.add(new Defconcept(concept));
      }
    }
    for (Iterator riter = newKB._roles.iterator(); riter.hasNext();) {
      Symbol role = (Symbol)riter.next();
      if (!roles.contains(role)) { // not yet defined
        System.err.println("Warning: role not defined: " + role);
        newKB.add(new Defrole(role));
      }
    }
    return newKB;
  }

  
  // printing
  // all methods return the abbreviation map
  
  public Map print() {
    return printLISP();
  }
  
  public Map printXML() {
    PrintWriter out = new PrintWriter(System.out);
    return printXML(out);
    // don't close System.out!!!
  }
  
  public Map printXML(PrintWriter out) {
    return print(new XMLSyntax(this), out);
  }
  
  public Map printLISP() {
    PrintWriter out = new PrintWriter(System.out);
    return printLISP(out);
    // don't close System.out!!!
  }
  
  public Map printLISP(PrintWriter out) {
    return print(new Syntax(this), out);
  }
  
  public Map print(Syntax syntax, PrintWriter out) {
    if (_abbrevs == null) 
      // we only use abbreviations when printing whole kbs!
      _abbrevs = new HashMap();
    if (syntax.XML()) {
      // out.... XML header
      out.println("<KNOWLEDGEBASE>");
    } else if (syntax.DOT()) {
      out.println("digraph ontology {");
      out.println("node [fontsize=11,fontname=Arial]"); // these should be parameters!
      out.println("edge [fontsize=11,fontname=Arial]");
    } else if (syntax.DAML()) {
      out.println("<rdf:RDF");
      out.println("  xmlns:rdf ='http://www.w3.org/1999/02/22-rdf-syntax-ns#'");
      out.println("  xmlns:rdfs='http://www.w3.org/2000/01/rdf-schema#'");
      out.println("  xmlns:daml='http://www.daml.org/2001/03/daml+oil#'");
      out.println(">\n");
    }
    // defconcept
    if (syntax.showDefconcept()) {
      for (Iterator iter = _assertions.iterator(); iter.hasNext();) {
        Assertion assertion = ((Assertion)iter.next());
        if (assertion instanceof Defconcept)
          assertion.print(syntax, out);
      }
    }
    // defrole
    if (syntax.showDefrole()) {
      for (Iterator iter = _assertions.iterator(); iter.hasNext();) {
        Assertion assertion = ((Assertion)iter.next());
        if (assertion instanceof Defrole)
          assertion.print(syntax, out);
      }
    }
    // other assertions
    for (Iterator iter = _assertions.iterator(); iter.hasNext();) {
      Assertion assertion = ((Assertion)iter.next());
      if (!(assertion instanceof Defconcept) && !(assertion instanceof Defrole))
        assertion.print(syntax, out);
    }
    if (syntax.XML() || syntax.LISP()) { // print namspaces as comment
      String prefix;
      if (syntax.XML()) {
        out.println("<!-- namespaces:");
        prefix = "  ";
      } else { // if (syntax.LISP()) 
        out.println(";;; namespaces:");
        prefix = ";;; ";
      }
      for (Iterator iter = _abbrevs.entrySet().iterator(); iter.hasNext();) {
        Map.Entry entry = (Map.Entry)iter.next();
        out.println(prefix + entry.getValue() + " = " + entry.getKey());
      }
      if (syntax.XML())
        out.println("-->");
    } else if (syntax.DOT()) { // create node with namespaces
      out.print("\"Namespaces:\\n");
      for (Iterator iter = _abbrevs.entrySet().iterator(); iter.hasNext();) {
        Map.Entry entry = (Map.Entry)iter.next();
        out.print(entry.getValue() + " = " + entry.getKey() + "\\l");
      }
      out.println("\" [shape=record,style=filled,color=lightgrey,fontsize=10,fontcolor=black]");
    }
    if (syntax.XML())
      out.println("</KNOWLEDGEBASE>");
    else if (syntax.DOT())
      out.println("}");
    else if (syntax.DAML())
      out.println("\n</rdf:RDF>");
    out.flush();
    return _abbrevs;
  }
  
  // auxiliaries for abbreviation handling (esp. invention)
  
  public String getAbbreviation(String namespace) {
    if (_abbrevs != null) {
      String abbrev = (String)_abbrevs.get(namespace);
      if (abbrev == null) {
        abbrev = inventAbbrev(namespace);
        _abbrevs.put(namespace, abbrev);
      }
      return abbrev;
    } else // no abbrevs used (when printing just an expression, not a kb)
      return null;
  }
  
  String inventAbbrev(String namespace) {
    String abbrev = null;
    if (namespace.indexOf("daml+oil#") != -1)
      abbrev = "daml";
    else if (namespace.indexOf("rdf-schema") != -1)
      abbrev = "rdfs";
    else if (namespace.indexOf("rdf-syntax") != -1)
      abbrev = "rdf";
    else if (namespace.indexOf("xmlschema") != -1)
      abbrev = "xmlschema";
    else { // split at :, /, \, and # to find last component
      String lastComponent = null;
      for (StringTokenizer tokens = new StringTokenizer(namespace, "/#:\\");
	   tokens.hasMoreTokens();) {
	lastComponent = tokens.nextToken();
      }
      if (lastComponent == null)
	return gensymAbbrev();
      else { // normalize, simplify etc.
	// remove www. and .com/.org etc.? ... !!!
	abbrev = normalize(lastComponent, 15);
      }
    }
    return makeUnique(abbrev);
  }

  String gensymAbbrev() { // invent new abbreviation
    String abbrev = "ns" + _gensymCounter;
    _gensymCounter++;
    if (_abbrevs.containsKey(abbrev))
      return gensymAbbrev();
    else
      return abbrev;
  }

  String makeUnique(String abbrev) {
    if (_abbrevs.containsKey(abbrev))
      return makeUnique(abbrev + "_");
    else
      return abbrev;
  }

  String normalize(String string, int max) {
    // replace not allowed characters by _; shorten to max characters
    StringBuffer buffer = new StringBuffer();
    if (Character.isDigit(string.charAt(0)))
      buffer.append("ns_"); // better: remove (usually version number) ... !!!
    int l = string.length();
    if (l > max) l = max;
    for (int i = 0; i < l; i++) {
      char c = string.charAt(i);
      if (Character.isLetterOrDigit(c)) // -, : etc. ??
	buffer.append(c);
      else
	buffer.append("_");
    }
    return buffer.toString();
  }


}
