package org.semanticweb.glp;

import java.util.*;
import java.lang.*;

import org.semanticweb.glp.PredicateGenerator;
import org.semanticweb.lp.Literal;
import org.semanticweb.lp.ProgramClause;
import org.semanticweb.lp.Term;
import org.semanticweb.lp.VariableTerm;

/**
 *  <p>
 *
 *  Title: Triple</p> <p>
 *
 *  Description: RDF Query and Inference Language</p> <p>
 *
 *  Copyright: Copyright (c) 2001</p> <p>
 *
 *  Company: Stanford University</p>
 *
 *@author     Stefan Decker
 *@created    January 23, 2002
 *@version    1.0
 */

public class GLPLloydToporTransform extends Object {

    /**
     *  Description of the Method
     *
     *@param  ruleHead       Description of the Parameter
     *@param  head           Description of the Parameter
     *@exception  Exception  Description of the Exception
     */
    public void generateRuleHead(GLogicFormula head, ArrayList ruleHead) throws Exception {
        if (head == null) {
            return;
        }
        else
                if (head instanceof GLPAtom) {
            ruleHead.add(((GLPAtom) head).toLPLiteral());
        }
        else
                if (head instanceof GLPAnd) {
            GLPAnd andFormula = (GLPAnd) head;
            generateRuleHead(andFormula.getLeft(), ruleHead);
            generateRuleHead(andFormula.getRight(), ruleHead);
        }
        else {
            throw new Exception("Unexpected Formula");
        }
    }


    /**
     *  Description of the Method
     *
     *@param  head  Description of the Parameter
     *@return       Description of the Return Value
     */
    public ArrayList generateRuleHead(GLogicFormula head) {
        ArrayList a = new ArrayList();
        try {
            generateRuleHead(head, a);
        } catch (Exception e) {
            System.out.println("Exception in generate Rule Head: " + e);
        }
        return a;
    }


    /**
     *  Description of the Method
     *
     *@param  rule  Description of the Parameter
     *@return       Description of the Return Value
     */
    public ArrayList translateFormula(GLPRule rule) {
        ArrayList rules = new ArrayList();
        HashMap h = new HashMap(7);
        for (Iterator i = rule.getVariables().iterator(); i.hasNext(); ) {
            Object v = i.next();
            h.put(v, v);
        }
        GLogicFormula copiedBody = null;
        if (rule.getBody() != null) {
            copiedBody = rule.getBody().copyFormula(h);
        }
        GLPRule newRule =
                new GLPRule(rule.getVariables(), rule.getHead(), copiedBody);
        translateFormula(newRule, rules);
        return rules;
    }


    /**
     *  Description of the Method
     *
     *@param  rule   Description of the Parameter
     *@param  rules  Description of the Parameter
     */
    public void translateFormula(GLPRule rule, ArrayList rules) {
        GLogicFormula body = rule.getBody();
        GLogicFormula head = rule.getHead();
        if(body== null) { // just facts
          rules.add(new ProgramClause(generateRuleHead(head), null));
          return;
        }
        ArrayList disjunctions = new ArrayList();
        ArrayList todo = new ArrayList();
        ArrayList ruleHead = new ArrayList();

        ArrayList variables = rule.getVariables();
        getDisjunction(true, body, disjunctions, todo);
        for (Iterator i = disjunctions.iterator(); i.hasNext(); ) {
            ArrayList conjunctions = new ArrayList();
            GLogicFormula disj = (GLogicFormula) i.next();
            getConjunction(true, disj, conjunctions, todo);
            ArrayList ruleBody = new ArrayList();
            for (Iterator k = conjunctions.iterator(); k.hasNext(); ) {
                GLogicFormula conj = (GLogicFormula) k.next();
                if (conj instanceof GLPAtom) {
                    ruleBody.add(((GLPAtom) conj).toLPLiteral());
                }
                else if ((conj instanceof GLPNot) && (((GLPNot) conj).isLiteral())) {
                    ruleBody.add(((GLPNot) conj).toLPLiteral());
                }
                else {
                    HashSet free = new HashSet();
                    HashSet bound = new HashSet();
                    conj.freeBoundVariables(free, bound);
                    Term[] lpArguments = new Term[free.size()];
                    ArrayList glpArguments = new ArrayList(free);
                    String varName;
                    ArrayList newVariables = new ArrayList();
                    int index = 0;
                    for (Iterator l = free.iterator(); l.hasNext(); ) {
                        GLPTermVariable variable = (GLPTermVariable) l.next();
                        newVariables.add(variable);
                        lpArguments[index] = (VariableTerm) variable.toLPTerm();
                        index++;
                    }
                    String newPredicateName = PredicateGenerator.getNewPredicateSymbol();
                    Literal lpAtom = new Literal(newPredicateName, lpArguments);
                    GLPAtom glpAtom = new GLPAtom(newPredicateName, glpArguments);
                    ruleBody.add(lpAtom);
                    todo.add(new GLPRule(newVariables, glpAtom, conj));
                }
            }
            rules.add(new ProgramClause(generateRuleHead(head), ruleBody));
        }
        for (Iterator i = todo.iterator(); i.hasNext(); ) {
            translateFormula((GLPRule) i.next(), rules);
        }
    }


    /**
     *  Gets the conjunction attribute of the GLPLloydToporTransform class
     *
     *@param  pol      Description of the Parameter
     *@param  formula  Description of the Parameter
     *@param  conj     Description of the Parameter
     *@param  todo     Description of the Parameter
     */
    void getConjunction(boolean pol, GLogicFormula formula, ArrayList conj, ArrayList todo) {

        if (formula instanceof GLPNot) {
            getConjunction(!pol, ((GLPNot) formula).getFormula(), conj, todo);
        }
        else if (pol && (formula instanceof GLPAnd)) {
            getConjunction(pol, ((GLPAnd) formula).getLeft(), conj, todo);
            getConjunction(pol, ((GLPAnd) formula).getRight(), conj, todo);
        }
        else if (!pol && (formula instanceof GLPOr)) {
            getConjunction(pol, ((GLPOr) formula).getLeft(), conj, todo);
            getConjunction(pol, ((GLPOr) formula).getRight(), conj, todo);
        }
        else if (!pol && (formula instanceof GLPImpliedBy)) {
            getConjunction(!pol, ((GLPImpliedBy) formula).getRight(), conj, todo);
            getConjunction(pol, ((GLPImpliedBy) formula).getLeft(), conj, todo);
        }
        else if (!pol && (formula instanceof GLPImplies)) {
            getConjunction(!pol, ((GLPImplies) formula).getLeft(), conj, todo);
            getConjunction(pol, ((GLPImplies) formula).getRight(), conj, todo);
        }
        else if (pol && (formula instanceof GLPEquiv)) {
            HashMap h = new HashMap(7);
            HashMap h1 = new HashMap(7);
            GLPImpliedBy nf1 = new GLPImpliedBy(((GLPEquiv) formula).getLeft(), ((GLPEquiv) formula).getRight());
            GLPImplies nf2 = new GLPImplies(((GLPEquiv) formula).getLeft().copyFormula(h), ((GLPEquiv) formula).getRight().copyFormula(h1));
            getConjunction(pol, nf1, conj, todo);
            getConjunction(pol, nf2, conj, todo);
        }
        else if (!pol && (formula instanceof GLPExists)) {
            HashSet free = new HashSet();
            formula.freeBoundVariables(free, new HashSet());
            ArrayList arguments = new ArrayList(free);
            String p = PredicateGenerator.getNewPredicateSymbol();
            GLPAtom atom = new GLPAtom(p, arguments);
            GLPNot not = new GLPNot(atom);
            conj.add(not);
            todo.add(new GLPRule(atom, formula));
        }
        else if (pol && (formula instanceof GLPExists)) {
            getConjunction(pol, ((GLPExists) formula).getFormula(), conj, todo);
        }
        else if (formula instanceof GLPForall) {
            GLPNot not = new GLPNot(((GLPForall) formula).getFormula());
            GLPExists newF = new GLPExists(((GLPForall) formula).getVariables(), not);
            getConjunction(!pol, newF, conj, todo);
        }
        else if (pol) {
            conj.add(formula);
        }
        else if (!pol) {
            GLPNot not = new GLPNot(formula);
            conj.add(not);
        }
    }


    /**
     *  Gets the disjunction attribute of the GLPLloydToporTransform class
     *
     *@param  pol      Description of the Parameter
     *@param  formula  Description of the Parameter
     *@param  todo     Description of the Parameter
     *@param  disj     Description of the Parameter
     */
    final void getDisjunction(boolean pol, GLogicFormula formula, ArrayList disj, ArrayList todo) {
        if (formula instanceof GLPNot) {
            getDisjunction(!pol, ((GLPNot) formula).getFormula(), disj, todo);
        }
        else if (pol && (formula instanceof GLPOr)) {
            getDisjunction(pol, ((GLPOr) formula).getLeft(), disj, todo);
            getDisjunction(pol, ((GLPOr) formula).getRight(), disj, todo);
        }
        else if (!pol && (formula instanceof GLPAnd)) {
            getDisjunction(pol, ((GLPAnd) formula).getLeft(), disj, todo);
            getDisjunction(pol, ((GLPAnd) formula).getRight(), disj, todo);
        }
        else if (pol && (formula instanceof GLPImpliedBy)) {
            getDisjunction(pol, ((GLPImpliedBy) formula).getLeft(), disj, todo);
            getDisjunction(!pol, ((GLPImpliedBy) formula).getRight(), disj, todo);
        }
        else if (pol && (formula instanceof GLPImplies)) {
            getDisjunction(!pol, ((GLPImplies) formula).getLeft(), disj, todo);
            getDisjunction(pol, ((GLPImplies) formula).getRight(), disj, todo);
        }
        else if (!pol && (formula instanceof GLPEquiv)) {
            HashMap h = new HashMap(7);
            HashMap h1 = new HashMap(7);
            GLPImpliedBy nf1 = new GLPImpliedBy(((GLPEquiv) formula).getLeft(), ((GLPEquiv) formula).getRight());
            GLPImplies nf2 = new GLPImplies(((GLPEquiv) formula).getLeft().copyFormula(h), ((GLPEquiv) formula).getRight().copyFormula(h1));
            getDisjunction(pol, nf1, disj, todo);
            getDisjunction(pol, nf2, disj, todo);
        }
        else if (!pol && (formula instanceof GLPExists)) {
            HashSet free = new HashSet(10);
            HashSet bound = new HashSet(10);
            formula.freeBoundVariables(free, bound);
            ArrayList arguments = new ArrayList(free);
            String p = PredicateGenerator.getNewPredicateSymbol();
            GLPAtom atom = new GLPAtom(p, arguments);
            GLPNot not = new GLPNot(atom);
            disj.add(not);
            todo.add(new GLPRule(atom, formula));
        }

        else if (pol && (formula instanceof GLPExists)) {
            getDisjunction(pol, ((GLPExists) formula).getFormula(), disj, todo);
        }
        else if (formula instanceof GLPForall) {
            GLPNot not = new GLPNot(((GLPForall) formula).getFormula());
            GLPExists newF = new GLPExists(((GLPForall) formula).getVariables(), not);
            getDisjunction(!pol, newF, disj, todo);
        }
        else if (!pol) {
            GLPNot not = new GLPNot(formula);
            disj.add(not);
        }
        else {
            disj.add(formula);
        }
    }


    /**
     *  The main program for the GLPLloydToporTransform class
     *
     *@param  argv  The command line arguments
     */
    public static void main(String[] argv) {
        ArrayList term1 = new ArrayList();
        term1.add(new GLPTermVariable("X1"));
        term1.add(new GLPTermConstant("a1"));
        GLPAtom at1 = new GLPAtom("p1", term1);

        ArrayList term2 = new ArrayList();
        term2.add(new GLPTermVariable("X2"));
        term2.add(new GLPTermConstant("a2"));
        GLPAtom at2 = new GLPAtom("p2", term2);

        ArrayList term3 = new ArrayList();
        term3.add(new GLPTermVariable("X1"));
        term3.add(new GLPTermVariable("X2"));
        GLPAtom at3 = new GLPAtom("p3", term3);

        GLPOr orf = new GLPOr(at1, at2);
        ArrayList variables = new ArrayList();
        variables.add(new GLPTermVariable("X1"));
        variables.add(new GLPTermVariable("X2"));
        GLPRule r = new GLPRule(variables, at3, orf);

        GLPLloydToporTransform lt = new GLPLloydToporTransform();
        ArrayList rules = lt.translateFormula(r);
        System.out.println("Original rule:" + r);
        System.out.println("Resulting rules ");
        for (Iterator i = rules.iterator(); i.hasNext(); ) {
            System.out.println(i.next().toString());
        }
         {
            ArrayList term4 = new ArrayList();
            term4.add(new GLPTermVariable("X"));
            term4.add(new GLPTermConstant("Y"));
            GLPAtom at4 = new GLPAtom("subset", term4);

            ArrayList term5 = new ArrayList();
            term5.add(new GLPTermVariable("Z"));
            term5.add(new GLPTermVariable("X"));
            GLPAtom at5 = new GLPAtom("in", term5);

            ArrayList term6 = new ArrayList();
            term6.add(new GLPTermVariable("Z"));
            term6.add(new GLPTermVariable("Y"));
            GLPAtom at6 = new GLPAtom("in", term6);
            GLPImplies f1 = new GLPImplies(at5, at6);
            ArrayList variables1 = new ArrayList();
            variables1.add(new GLPTermVariable("Z"));
            GLPForall f2 = new GLPForall(variables1, f1);
            ArrayList variables2 = new ArrayList();
            variables2.add(new GLPTermVariable("X"));
            variables2.add(new GLPTermVariable("Y"));
            GLPRule r1 = new GLPRule(variables2, at4, f2);
            System.out.println("Original rule:" + r1);
            System.out.println("Resulting rules ");
            ArrayList rules2 = lt.translateFormula(r1);
            for (Iterator i = rules2.iterator(); i.hasNext(); ) {
                System.out.println(i.next().toString());
            }
        }
         {
            ArrayList term4 = new ArrayList();
            term4.add(new GLPTermVariable("X"));
            term4.add(new GLPTermVariable("Y"));
            GLPAtom at4 = new GLPAtom("subset", term4);

            ArrayList term5 = new ArrayList();
            term5.add(new GLPTermVariable("Z"));
            term5.add(new GLPTermVariable("X"));
            GLPAtom at5 = new GLPAtom("in", term5);

            ArrayList term6 = new ArrayList();
            term6.add(new GLPTermVariable("Z"));
            term6.add(new GLPTermVariable("Y"));
            GLPAtom at6 = new GLPAtom("in", term6);
            GLPEquiv f1 = new GLPEquiv(at5, at6);
            ArrayList variables1 = new ArrayList();
            variables1.add(new GLPTermVariable("Z"));
            GLPForall f2 = new GLPForall(variables1, f1);
            ArrayList variables2 = new ArrayList();
            variables2.add(new GLPTermVariable("X"));
            variables2.add(new GLPTermVariable("Y"));
            GLPRule r1 = new GLPRule(variables2, at4, f2);
            System.out.println("Original rule:" + r1);
            System.out.println("Resulting rules ");
            ArrayList rules2 = lt.translateFormula(r1);
            for (Iterator i = rules2.iterator(); i.hasNext(); ) {
                System.out.println(i.next().toString());
            }
        }
         {
            ArrayList term4 = new ArrayList();
            term4.add(new GLPTermVariable("X"));
            GLPAtom at4 = new GLPAtom("p", term4);

            ArrayList term5 = new ArrayList();
            term5.add(new GLPTermVariable("Z"));
            term5.add(new GLPTermVariable("X"));
            GLPAtom at5 = new GLPAtom("in", term5);

            ArrayList variables1 = new ArrayList();
            variables1.add(new GLPTermVariable("Z"));
            GLPExists f1 = new GLPExists(variables1, at5);

            ArrayList term6 = new ArrayList();
            term6.add(new GLPTermVariable("Z"));
            term6.add(new GLPTermVariable("X"));
            GLPAtom at6 = new GLPAtom("in", term6);
            GLPExists f2 = new GLPExists(variables1, at6);
            GLPAnd f3 = new GLPAnd(f1, f2);

            ArrayList variables2 = new ArrayList();
            variables2.add(new GLPTermVariable("X"));
            GLPRule r1 = new GLPRule(variables2, at4, f3);
            System.out.println("Original rule:" + r1);
            System.out.println("Resulting rules ");
            ArrayList rules2 = lt.translateFormula(r1);
            for (Iterator i = rules2.iterator(); i.hasNext(); ) {
                System.out.println(i.next().toString());
            }
        }
         {
            ArrayList term14 = new ArrayList();
            term14.add(new GLPTermVariable("X"));
            term14.add(new GLPTermVariable("Y"));
            GLPAtom at14 = new GLPAtom("subset", term14);

            ArrayList term24 = new ArrayList();
            term24.add(new GLPTermVariable("X"));
            term24.add(new GLPTermVariable("Y"));
            GLPAtom at24 = new GLPAtom("newsubset", term24);

            GLPAnd at4 = new GLPAnd(at14, at24);

            ArrayList term5 = new ArrayList();
            term5.add(new GLPTermVariable("Z"));
            term5.add(new GLPTermVariable("X"));
            GLPAtom at5 = new GLPAtom("in", term5);

            ArrayList term6 = new ArrayList();
            term6.add(new GLPTermVariable("Z"));
            term6.add(new GLPTermVariable("Y"));
            GLPAtom at6 = new GLPAtom("in", term6);
            GLPImplies f1 = new GLPImplies(at5, at6);
            ArrayList variables1 = new ArrayList();
            variables1.add(new GLPTermVariable("Z"));
            GLPForall f2 = new GLPForall(variables1, f1);
            ArrayList variables2 = new ArrayList();
            variables2.add(new GLPTermVariable("X"));
            variables2.add(new GLPTermVariable("Y"));
            GLPRule r1 = new GLPRule(variables2, at4, f2);
            System.out.println("Original rule:" + r1);
            System.out.println("Resulting rules ");
            ArrayList rules2 = lt.translateFormula(r1);
            for (Iterator i = rules2.iterator(); i.hasNext(); ) {
                System.out.println(i.next().toString());
            }
        }

    }
}

