package org.semanticweb.triple.language;

import java.util.*;
import org.semanticweb.glp.*;

public class Clause extends TripleExpression {

  ArrayList variables = new ArrayList(); // of Variable
  Formula head;
  Formula body;
  ArrayList modelVariables = new ArrayList(); // of Variable
  Term model;

  public void putVariables(Term variable) { variables.add(variable); }
  public void putHead(Formula head) { this.head = head; }
  public void putBody(Formula body) { this.body = body; }
  public void putModelVariables(Term variable) { modelVariables.add(variable); }
  public void putModel(Term model) { this.model = model; }

  public ArrayList getVariables() { return variables; }
  public Formula getHead() { return head; }
  public Formula getBody() { return body; }
  public ArrayList getModelVariables() { return modelVariables; }
  public Term getModel() { return model; }

  public String toTripleString(boolean withModel) {
    return toString(new TripleSyntax(), withModel);
  }

  public String toString(Syntax syntax) {
    return toString(syntax, true);
  }

  public String toString(Syntax syntax, boolean withModel) {
    StringBuffer buffer = new StringBuffer();
    if (withModel && model != null) {
      if (modelVariables != null && modelVariables.size() != 0) {
        buffer.append("forall ");
        appendList(modelVariables, buffer, syntax, ",");
        buffer.append(" ");
      }
      buffer.append("@");
      buffer.append(model.toString(syntax));
      buffer.append(" { ");
    }
    if (variables != null && variables.size() != 0) {
      buffer.append("forall ");
      appendList(variables, buffer, syntax, ",");
      buffer.append(" ");
    }
    if (head != null && body != null) { // rule
      buffer.append(head.toString(syntax));
      buffer.append(syntax.getIf());
      buffer.append(body.toString(syntax));
    } else if (head != null) { // fact
      buffer.append(head.toString(syntax));
    } else { // query
      buffer.append(syntax.getQueryIf());
      if (body != null)
        buffer.append(body.toString(syntax));
      /* not needed any more
      if (prolog && prologQuery) {
	buffer.append(", writeq(");
	buffer.append(body.toString(syntax));
	buffer.append("), write('.'), nl, fail");
      }
      */
    }
    buffer.append(".");
    if (withModel && model != null) {
      buffer.append(" }");
    }
    return buffer.toString();
  }

  public GLPClause toGLP() {
    GLogicFormula glpHead = null;
    GLogicFormula glpBody = null;
    ArrayList additionalGlpFormulae = new ArrayList(); // from model expression
    GLPTerm glpModel = model.toGLP(new GLPTermVariable(),
      additionalGlpFormulae);
      // this is not correct: it moves the add. formulae into the body!!!
      // we probably should forbid this anyway ...;
      // what about outer models? should they be used instead of
      // the anonymous variable??
    if (head != null)
      glpHead = (GLogicFormula)head.toGLP(glpModel);
      // hmmm, is this correct? what about additional guard-like
      // head conjuncts; at the moment, only triples are generated,
      // which belong in the head, so everything is fine
    if (body != null)
      glpBody = (GLogicFormula)body.toGLP(glpModel);
    glpBody = And.makeGLPConjunction(additionalGlpFormulae, glpBody);
    ArrayList glpVars = new ArrayList();
    addAsGLPVariables(modelVariables, glpVars);
    addAsGLPVariables(variables, glpVars);
    return new GLPRule(glpVars, glpHead, glpBody);
  }

  // auxiliaries
  
  static void addAsGLPVariables(ArrayList variables, ArrayList glpVariables) {
    for (Iterator varIter = variables.iterator(); varIter.hasNext();) {
      Variable var = (Variable)varIter.next();
      glpVariables.add(new GLPTermVariable(var.getLabel()));
    }
  }

  
}
