/*
 * PrintCallback.java
 *
 * Created on January 25, 2002, 4:20 PM
 */

package org.semanticweb.triple.parser;

import java.util.*;
import java.io.*;
import org.semanticweb.triple.language.*;
import org.semanticweb.glp.*;
import org.semanticweb.lp.*;


/**
 * Debugging implementation of Callback.
 * @author  Michael Sintek
 * @version 
 */
public class PrintCallback implements Callback {

  PrintWriter writer;
  
  public PrintCallback() {
    this.writer = new PrintWriter(System.out);
  }
  
  public PrintCallback(PrintWriter writer) {
    this.writer = writer;
  }
  
  public void nextClause (LTerm lclause, Context context) {
    LTerm lmodel = context.getModel();
    writer.println("HIL: " + lclause + " @ " + lmodel); // hilog representation
    try {
      // 1. native TRIPLE representation
      Clause clause = TripleTransform.transformClause(lclause, lmodel);
      writer.println("TRP: " + clause);
      // 2. GLP
      GLPClause glpClause = clause.toGLP();
      writer.println("GLP: " + glpClause);
      // 3. LP (Lloyd Topor application)
      GLPLloydToporTransform lt = new GLPLloydToporTransform();
      ArrayList lpRules = lt.translateFormula((GLPRule)glpClause);
      for (Iterator i = lpRules.iterator(); i.hasNext();) {
        ProgramClause lpClause = (ProgramClause)i.next();
        writer.println(" LP: " + lpClause);
      }
      writer.flush();
    } catch (Exception e) {
      writer.println("???: " + e);
      e.printStackTrace(); // activate to see line numbers!!
    }
    writer.println();
  }
    
}
